# Copyright 2013-2022 Reahl Software Services (Pty) Ltd. All rights reserved.
#
#    This file is part of Reahl.
#
#    Reahl is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation; version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.


import inspect
try:
    from functools import cached_property
except ImportError:
    from cached_property import cached_property
from reahl.component.decorators import  deprecated


try:
    import contextvars
    execution_context_var = contextvars.ContextVar('reahl.component.context.ExecutionContext')
except:
    execution_context_var = None


class NoContextFound(Exception):
    pass




class ExecutionContext:
    """Most code execute "in the scope of" some ExecutionContext. Such code can obtain
       the current ExecutionContext by calling ExecutionContext.get_context. The ExecutionContext
       of code comprises of: the current Configuration for all components, the current UserSession,
       and a SystemControl.
    
       .. attribute:: config
       
       .. attribute:: session
       
       .. attribute:: system_control
    """
    use_context_var = False
    
    @classmethod
    def for_config_directory(cls, config_directory):
        from reahl.component.config import StoredConfiguration # Here, to avoid circular dependency
        
        config = StoredConfiguration(config_directory)
        config.configure()
        new_context = cls()
        new_context.config = config
        return new_context

    @classmethod
    def get_context_id(cls):
        """Returns a unique, hashable identifier identifying the current call context."""
        return cls.get_context().id


    @classmethod
    def get_context(cls):
        """Returns the current call context, or raises :class:`NoContextFound` if there is none."""
        if not cls.use_context_var:
            return cls.get_context_using_frames()
        try:
            context = execution_context_var.get()
        except LookupError:
            context = None
        if not context:
            raise NoContextFound('No %s is active in the call stack' % cls)

        return context

    @classmethod
    def get_context_using_frames(cls):
        """Returns the current call context, or raises :class:`NoContextFound` if there is none."""
        context = None
        f = inspect.currentframe()
        while (not context) and f:
            context = f.f_locals.get('__reahl_context__', None)
            to_delete = f
            f = f.f_back
            del to_delete
        if not context:
            raise NoContextFound('No %s is active in the call stack' % cls)
        return context

    def __init__(self, name=None, parent_context=None):
        self.name = name
        try:
            self.parent_context = parent_context or self.get_context()
        except NoContextFound:
            self.parent_context = None
        self.id = (self.parent_context.id if isinstance(self.parent_context, ExecutionContext) else id(self))

    def copy(self):
        context = self.__class__(name='ExecutionContext.copy()', parent_context=self)
        for name in self.__dict__:
            if name not in ['id', 'name','parent_context']:
                context.__setattr__(name, self.__dict__[name])
        return context

    def __enter__(self):
        self.install_with_context_vars_or_frames()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.uninstall()

    def install_with_context_vars_or_frames(self, stop=None):
        if not self.use_context_var:
            return self.install_using_frames(stop=stop)
        try:
            if execution_context_var.get() == self:
                raise Exception('Context already installed')
        except LookupError:
            pass
        self.ctx_token = execution_context_var.set(self)
        return self


    @deprecated('Use `with ExecutionContext(): ...` instead, in order to exit the context explicitly as well.', '6.1')
    def install(self, stop=None):
        """Installs the ExecutionContext in the current scope so that it will be found by code called after it is installed.

           .. versionchanged:: 6.1
              Deprecated install. Use ExecutionContext as ContextManager instead.

        """
        return self.install_with_context_vars_or_frames(stop=stop)

    def uninstall(self):
        if not self.use_context_var:
            return
        execution_context_var.reset(self.ctx_token)

    def install_using_frames(self, stop=None):
        f = inspect.currentframe()
        if not stop:
            def stop(ff):
                return not isinstance(ff.f_locals.get('self', None), ExecutionContext)
        while not stop(f):
            calling_frame = f.f_back
            del f
            f = calling_frame
        calling_frame.f_locals['__reahl_context__'] = self
        del calling_frame
        return self

    @cached_property
    def interface_locale(self):
        """Returns a string identifying the current locale."""
        session = getattr(self, 'session', None)
        if not session:
            return 'en_gb'
        return self.session.get_interface_locale()

    def __getattr__(self, name):
        raise AttributeError('%s has no attribute \'%s\'' % (self, name))

    def __str__(self):
        if self.name:
            return '%s named \'%s\'' % (self.__class__.__name__, self.name)
        return super().__str__()

        



# Copyright 2013-2022 Reahl Software Services (Pty) Ltd. All rights reserved.
#
#    This file is part of Reahl.
#
#    Reahl is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation; version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Support for database schema migration."""

import logging
import warnings
import inspect
import traceback

from reahl.component.exceptions import ProgrammerError
from reahl.component.eggs import DependencyGraph, DependencyCluster


class MigrationPlan:
    def __init__(self, root_egg, orm_control):
        self.root_egg = root_egg
        self.orm_control = orm_control
        self.version_graph = None
        self.cluster_graph = None
        self.schedules = None

    def do_planning(self):
        self.version_graph = self.create_version_graph_for(self.root_egg, self.orm_control)
        self.version_graph.topological_sort() # to detect circular dependencies

        self.cluster_graph = self.create_cluster_graph(self.version_graph)
        self.all_clusters_in_smallest_first_topological_order = list(reversed(list(self.cluster_graph.topological_sort())))
        self.schedules = self.create_schedules_for_clusters(self.all_clusters_in_smallest_first_topological_order)

    @classmethod
    def create_version_graph_for(cls, egg, orm_control):
        graph = {}
        versions = egg.get_versions()
        for version in versions:
            cls.discover_version_graph_for(version, graph, orm_control)
        return DependencyGraph(graph)

    @classmethod
    def discover_version_graph_for(cls, version, graph, orm_control):
        if not version:
            return
        if version not in graph and not version.is_up_to_date(orm_control):
            cls.discover_version_graph_for(version.get_previous_version(), graph, orm_control)
            all_versions = [dependency.get_best_version() for dependency in version.get_dependencies()
                            if dependency.distribution and dependency.is_component]
            children = graph[version] = [v for v in all_versions if not v.is_up_to_date(orm_control)]
            for v in children:
                cls.discover_version_graph_for(v, graph, orm_control)
                cls.discover_version_graph_for(v.get_previous_version(), graph, orm_control)

    @classmethod
    def create_cluster_graph(cls, version_graph):
        clusters = [DependencyCluster(root, contents) for root, contents in version_graph.find_components()]
        return DependencyGraph.from_vertices(clusters, lambda c: c.get_dependencies(clusters))

    def execute(self):
        for schedule in self.schedules:
            schedule.execute_all()
        if self.schedules:
            self.orm_control.prune_schemas_to_only(self.schedules[-1].cluster.versions)

    def explain(self):
        if self.version_graph:
            print('Rendering version graph to: versions.svg')
            self.version_graph.render('versions')
        else: 
            print('Could not compute version graph')
        if self.cluster_graph:
            print('Rendering cluster graph to: clusters.svg')
            self.cluster_graph.render('clusters')
        else: 
            print('Could not compute cluster graph')
        if self.schedules is not None:
            def find_schedules(schedules):
                all_schedules = schedules[:]
                for schedule in schedules:
                    all_schedules.extend(find_schedules(schedule.nested_schedules))
                return all_schedules
            schedule_graph = DependencyGraph.from_vertices(find_schedules(self.schedules), lambda r: r.nested_schedules)
            print('Rendering schedule graph to: schedules.svg')
            schedule_graph.render('schedules')
        else: 
            print('Could not compute schedule graph')
        
    def create_schedules_for_clusters(self, clusters_in_smallest_first_topological_order):
        schedules = []
        for cluster in clusters_in_smallest_first_topological_order:
            if not cluster.visited:
                cluster_children = cluster.get_dependencies(clusters_in_smallest_first_topological_order)
                next_cluster = self.find_highest_parent_of(cluster_children, clusters_in_smallest_first_topological_order) if cluster_children else cluster
                next_cluster.visited = True 
                schedules.append(self.create_migration_schedule_for(next_cluster, clusters_in_smallest_first_topological_order))
        return schedules

    def find_highest_parent_of(self, clusters, clusters_in_smallest_first_topological_order):
        immediate_ancestors = [a for a in clusters_in_smallest_first_topological_order
                                if any([c in a.get_dependencies(clusters_in_smallest_first_topological_order) for c in clusters])]
        return immediate_ancestors[-1]

    def create_migration_schedule_for(self, cluster, clusters_in_smallest_first_topological_order):
        logging.getLogger(__name__).debug('Creating MigrationSchedule for cluster %s' % cluster)
        schedule = MigrationSchedule(self.orm_control, cluster, self.all_clusters_in_smallest_first_topological_order) 
        schedule.schedule_migrations()
        unhandled_decendants_in_smallest_first_topological_order = [c for c in clusters_in_smallest_first_topological_order
                                                                    if (clusters_in_smallest_first_topological_order.index(c) < clusters_in_smallest_first_topological_order.index(cluster))
                                                                        and not c.visited]
        if unhandled_decendants_in_smallest_first_topological_order:
            logging.getLogger(__name__).debug('Adding nested MigrationSchedules for clusters: %s' % unhandled_decendants_in_smallest_first_topological_order)
        for nested_schedule in self.create_schedules_for_clusters(unhandled_decendants_in_smallest_first_topological_order):
            schedule.add_nested(nested_schedule)
        if unhandled_decendants_in_smallest_first_topological_order:
            logging.getLogger(__name__).debug('Done adding nested MigrationSchedules for clusters: %s' % unhandled_decendants_in_smallest_first_topological_order)
        return schedule


class NoMigrationScheduleFound(Exception):
    def __init__(self, cluster):
        super().__init__('Could not find nested schedules for %s' % cluster)


class ExceptionDuringMigration(Exception):
    def __init__(self, scheduling_context):
        super(ExceptionDuringMigration, self).__init__()
        self.scheduling_context = scheduling_context
    def __str__(self):
        message = super(ExceptionDuringMigration, self).__str__()
        formatted_context = traceback.format_list([(frame_info.filename, frame_info.lineno, frame_info.function, frame_info.code_context[frame_info.index])
                                                    for frame_info in self.scheduling_context])
        return '%s\n\n%s\n\n%s' % (message, 'The above Exception happened for the migration that was scheduled here:', ''.join(formatted_context))


class MigrationSchedule:
    """A schedule stating in which order migration operations should be performed to bring the database up to date with the given DependencyCluster"""
    def __init__(self, orm_control, cluster, all_clusters):
        self.cluster = cluster
        self.orm_control = orm_control
        self.phases_in_order = ['drop_pk', 'pre_alter', 'alter', 'create_pk', 'indexes', 'data', 'create_fk', 'cleanup']
        self.phases = dict([(i, []) for i in self.phases_in_order])
        self.before_nesting_phase = []
        self.last_phase = []
        self.after_nested_schedule_phases = {}
        self.nested_schedules = []
        self.all_clusters = all_clusters
        self.current_drop_fk_phase = self.before_nesting_phase

    def __repr__(self):
        return '<%s %s>' % (self.__class__.__name__, self.cluster)

    def get_previous_cluster_for_version(self, version):
        clusters_containing_previous_version = [c for c in self.all_clusters
                                                if any([v.is_previous_version_of(version) for v in c.versions])]
        previous_cluster = None
        if clusters_containing_previous_version:
            assert len(clusters_containing_previous_version) == 1
            [previous_cluster] = clusters_containing_previous_version

        return previous_cluster

    def schedule_migrations(self):
        for version in self.cluster.get_versions_biggest_first():
            previous_cluster = self.get_previous_cluster_for_version(version)

            if previous_cluster and not previous_cluster.visited:
                self.current_drop_fk_phase = self.after_nested_phase_for(previous_cluster)
            else:
                self.current_drop_fk_phase = self.before_nesting_phase

            for migration_class in version.get_migration_classes():
                migration = migration_class(self)
                logging.getLogger(__name__).debug('Scheduling Migrations for migration %s in cluster %s' % (migration_class, self.cluster))
                migration.schedule_upgrades()
            UpdateSchemaVersion(self, version).schedule_upgrades()

    def add_nested(self, schedule):
        if any([s.cluster is schedule.cluster for s in self.nested_schedules]):
            raise ProgrammerError('Trying to add %s, but there is already a nested schedule for %s' % (schedule, schedule.cluster))
        self.nested_schedules.append(schedule)
        
    def after_nested_phase_for(self, cluster):
        return self.after_nested_schedule_phases.setdefault(cluster, [])

    def schedule(self, phase, scheduling_migration, scheduling_context, to_call, *args, **kwargs):
        if phase == 'drop_fk':
            self.current_drop_fk_phase.append((to_call, scheduling_migration, scheduling_context, args, kwargs))
        else:
            try:
                self.phases[phase].append((to_call, scheduling_migration, scheduling_context, args, kwargs))
            except KeyError as e:
                raise ProgrammerError('A phase with name<%s> does not exist.' % phase)

    def execute(self, scheduled_changes):
        for to_call, scheduling_migration, scheduling_context, args, kwargs in scheduled_changes:
            logging.getLogger(__name__).info('Executing for %s in %s: %s(%s, %s)' % (scheduling_migration, self.cluster, to_call.__name__, args, kwargs))
            try:
                to_call(*args, **kwargs)
            except Exception as e:
                raise ExceptionDuringMigration(scheduling_context)

    def execute_nested_schedules(self):
        for schedule in self.nested_schedules:
            schedule.execute_all()
            self.execute(self.after_nested_schedule_phases[schedule.cluster])

    def execute_all(self):
        self.execute(self.before_nesting_phase)
        self.execute_nested_schedules()
        for phase in self.phases_in_order:
            self.execute(self.phases[phase])


class Migration:
    """Represents one logical change that can be made to an existing database schema.
    
       You should extend this class to build your own domain-specific database schema migrations.
       
       Never use code imported from your component in a Migration, since Migration code is kept around in
       future versions of a component and may be run to migrate a schema with different versions of the code in your component.
    """

    def __init__(self, migration_schedule):
        self.migration_schedule = migration_schedule

    @property
    def orm_control(self):
        return self.migration_schedule.orm_control

    def schedule(self, phase, to_call, *args, **kwargs):
        """Call this method to schedule a method call for execution later during the specified migration phase.

           Scheduled migrations are first collected from all components, then the calls scheduled for each defined
           phase are executed. All the calls scheduled in the same phase are executed in the order they were scheduled. 
           Phases are executed in the following order:

           'drop_fk', 'drop_pk', 'pre_alter', 'alter', 'create_pk', 'indexes', 'data', 'create_fk', 'cleanup'

           :param phase: The name of the phase to schedule this call.
           :param to_call: The method or function to call.
           :param args: The positional arguments to be passed in the call.
           :keyword kwargs: The keyword arguments to be passed in the call.
        """
        def get_scheduling_context():
            relevant_frames = []
            found_framework_code = False
            frames = iter(inspect.stack()[2:]) #remove this function from the stack
            while not found_framework_code:
                try:
                    frame = next(frames)
                    found_framework_code = frame.filename.endswith(__file__)
                    if (not found_framework_code) or frame.function == 'schedule_upgrades':
                        relevant_frames.append(frame)
                except StopIteration:
                    found_framework_code = True

            return list(reversed(relevant_frames))
        self.migration_schedule.schedule(phase, self, get_scheduling_context(), to_call, *args, **kwargs)

    def schedule_upgrades(self):
        """Override this method in a subclass in order to supply custom logic for changing the database schema. This
           method will be called for each of the applicable Migrations listed for all components. The order in which
           Migrations are scheduled is carefully crafted based on dependencies between versions.

           **Added in 2.1.2**: Supply custom upgrade logic by calling `self.schedule()`.
        """
        warnings.warn('Ignoring %s.schedule_upgrades(): it does not override schedule_upgrades() (method name typo perhaps?)' % self.__class__.__name__ , 
                      UserWarning, stacklevel=-1)

    @property
    def name(self):
        return '%s.%s' % (inspect.getmodule(self).__name__, self.__class__.__name__)

    def __str__(self):
        return self.name


class UpdateSchemaVersion(Migration):
    def __init__(self, migration_schedule, version):
        super().__init__(migration_schedule)
        self.version = version

    def schedule_upgrades(self):
        self.schedule('cleanup', self.migration_schedule.orm_control.set_schema_version_for, self.version)



# Copyright 2013-2022 Reahl Software Services (Pty) Ltd. All rights reserved.
#
#    This file is part of Reahl.
#
#    Reahl is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation; version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.



from reahl.stubble import EmptyStub
from reahl.tofu import Fixture, scenario, with_fixtures, set_up, tear_down
from reahl.component.context import ExecutionContext

class ContextImplementationFixture(Fixture):

    @set_up
    def record_original_implementation_setting(self):
        self.original_use_context_var = ExecutionContext.use_context_var

    @tear_down
    def restore_original_implementation_setting(self):
        ExecutionContext.use_context_var = self.original_use_context_var

    @scenario
    def use_frames(self):
        ExecutionContext.use_context_var = False

    @scenario
    def use_contextvars(self):
        ExecutionContext.use_context_var = True


@with_fixtures(ContextImplementationFixture)
def test_execution_context_basics(context_implementation_fixture):
    """An ExecutionContext is like a global variable for a particular call stack. To create an
       ExecutionContext for a call stack, use it in a with statement."""

    def do_something():
        return ExecutionContext.get_context()
    def do_high_level_something():
        return do_something()

    with ExecutionContext() as some_context:

        found_context = do_high_level_something()

        assert found_context is some_context


@with_fixtures(ContextImplementationFixture)
def test_execution_context_stacking(context_implementation_fixture):
    """When an ExecutionContext overrides a deeper one on the call stack, it will retain the same id."""
    with ExecutionContext() as some_context:

        def deeper_code():
            with ExecutionContext() as deeper_context:
                assert ExecutionContext.get_context_id() == some_context.id
                return deeper_context

        deeper_context = deeper_code()
        assert some_context is not deeper_context
        assert some_context.id == deeper_context.id


@with_fixtures(ContextImplementationFixture)
def test_contents(context_implementation_fixture):
    """A Session, Config or SystemControl may be set on the ExecutionContext."""
    some_context = ExecutionContext()

    session = EmptyStub()
    config = EmptyStub()
    system_control = EmptyStub()

    some_context.session = session
    some_context.config = config
    some_context.system_control = system_control

    assert some_context.session is session
    assert some_context.config is config
    assert some_context.system_control is system_control






# Copyright 2013-2022 Reahl Software Services (Pty) Ltd. All rights reserved.
#
#    This file is part of Reahl.
#
#    Reahl is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation; version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.


import pkg_resources
import re
import logging

from reahl.tofu import Fixture, set_up, temp_dir, expected
from reahl.tofu.pytestsupport import with_fixtures
from reahl.stubble import CallMonitor, EasterEgg, easter_egg

from reahl.component.eggs import ReahlEgg
from reahl.component.config import Configuration, StoredConfiguration, ConfigSetting, \
                                   ConfigurationException, EntryPointClassList, DeferredDefault


class ConfigWithFiles(Fixture):
    def new_config_dir(self):
        return temp_dir()
    
    def new_config_bootstrap_file(self):
        contents = """
reahlsystem.root_egg = '%s'
reahlsystem.connection_uri = None
reahlsystem.debug = False
""" % self.root_egg_name
        return self.new_config_file(filename='reahl.config.py', contents=contents)

    def new_root_egg_name(self):
        return easter_egg.as_requirement_string()

    def new_config_file_name(self):
        return 'some_file.py'
        
    def new_config_file(self, filename=None, contents=''):
        return self.config_dir.file_with(filename or self.config_file_name, contents)
        
    @set_up
    def set_up_easter_egg(self):
        self.config_bootstrap_file
        easter_egg.clear()
        easter_egg.stubbed_metadata['reahl-component.toml'] = 'metadata_version = "1.0.0"'
        ReahlEgg.clear_cache()

    def set_config_spec(self, egg, code_locator_string):
        egg.stubbed_metadata['reahl-component.toml'] = 'metadata_version = "1.0.0"\nconfiguration = "%s"' % code_locator_string



class ConfigWithSetting(Configuration):
    filename = 'config_file_for_this_egg.py'
    config_key = 'some_key'
    some_setting = ConfigSetting()


@with_fixtures(ConfigWithFiles)
def test_config_basics(config_with_files):
    """Config is specified per component, via its ReahlEgg interface, and read from its own config file."""

    fixture = config_with_files
    config_file = fixture.new_config_file(filename=ConfigWithSetting.filename, 
                                          contents='some_key.some_setting = 3')
    fixture.set_config_spec(easter_egg, 'reahl.component_dev.test_config:ConfigWithSetting')

    # Usually this happens inside other infrastructure, such as the implementation of reahl serve (see reahl-dev)
    config = StoredConfiguration(fixture.config_dir.name)
    config.configure()

    # The setting was read
    assert config.some_key.some_setting == 3 


@with_fixtures(ConfigWithFiles)
def test_missing_settings(config_with_files):
    """An exception is raised if setting do not have values set."""
    fixture = config_with_files
    
    fixture.set_config_spec(easter_egg, 'reahl.component_dev.test_config:ConfigWithSetting')

    # Usually this happens inside other infrastructure, such as the implementation of reahl serve (see reahl-dev)
    config = StoredConfiguration(fixture.config_dir.name)

    with expected(ConfigurationException):
        config.configure()


@with_fixtures(ConfigWithFiles)
def test_incorrect_settings(config_with_files):
    """An exception is raised when an attempt is made to set a setting that does not exist."""

    fixture = config_with_files
    config_file = fixture.new_config_file(filename=ConfigWithSetting.filename, 
                                          contents='some_key.some_wrong_name = 3')
    fixture.set_config_spec(easter_egg, 'reahl.component_dev.test_config:ConfigWithSetting')

    # Usually this happens inside other infrastructure, such as the implementation of reahl serve (see reahl-dev)
    config = StoredConfiguration(fixture.config_dir.name)

    with expected(ConfigurationException):
        config.configure()


@with_fixtures(ConfigWithFiles)
def test_incorrect_replacement_of_configuration(config_with_files):
    """An exception is raised when an attempt is made to replace a Configuration with another of the wrong type."""

    fixture = config_with_files
    config_file = fixture.new_config_file(filename=ConfigWithSetting.filename, 
                                          contents='from reahl.component.config import Configuration; some_key = Configuration()')
    fixture.set_config_spec(easter_egg, 'reahl.component_dev.test_config:ConfigWithSetting')

    # Usually this happens inside other infrastructure, such as the implementation of reahl serve (see reahl-dev)
    config = StoredConfiguration(fixture.config_dir.name)

    with expected(ConfigurationException):
        config.configure()


class ConfigWithDefaultedSetting(Configuration):
    filename = 'config_file_for_this_egg.py'
    config_key = 'some_key'
    some_setting = ConfigSetting(default='default value')


@with_fixtures(ConfigWithFiles)
def test_config_defaults(config_with_files):
    """If a default is specified for the ConfigSetting, it need not be set in the file."""

    fixture = config_with_files
    fixture.set_config_spec(easter_egg, 'reahl.component_dev.test_config:ConfigWithDefaultedSetting')

    # Usually this happens inside other infrastructure, such as the implementation of reahl serve (see reahl-dev)
    config = StoredConfiguration(fixture.config_dir.name)
    config.configure()

    # The setting was read
    assert config.some_key.some_setting == 'default value' 


class ConfigWithDangerousDefaultedSetting(Configuration):
    filename = 'config_file_for_this_egg.py'
    config_key = 'some_key'
    some_setting = ConfigSetting(default='default value', dangerous=True)


@with_fixtures(ConfigWithFiles)
def test_config_defaults_dangerous(config_with_files):
    """Defaults that are dangerous to leave at their at their settings can be marked as such. 
       This will result in a logged warning."""

    fixture = config_with_files
    fixture.set_config_spec(easter_egg, 'reahl.component_dev.test_config:ConfigWithDangerousDefaultedSetting')

    # Usually this happens inside other infrastructure, such as the implementation of reahl serve (see reahl-dev)
    config = StoredConfiguration(fixture.config_dir.name)
    with CallMonitor(logging.getLogger('reahl.component.config').warning) as monitor:
        config.configure()

    # A warning was issued with warn severity to the log
    logged_message = monitor.calls[0].args[0]
    message_regex = '^some_key.some_setting has been defaulted to a value not suitable for production use: "default value". You can set it in /.*/config_file_for_this_egg.py'
    assert re.match(message_regex, logged_message)

    # The default value is still used
    assert config.some_key.some_setting == 'default value' 


@with_fixtures(ConfigWithFiles)
def test_config_strict_checking(config_with_files):
    """When a Configuration is created with strict_checking=True, dangerous defaulted config is not allowed."""

    fixture = config_with_files
    fixture.set_config_spec(easter_egg, 'reahl.component_dev.test_config:ConfigWithDangerousDefaultedSetting')

    config = StoredConfiguration(fixture.config_dir.name, strict_checking=True)
    with expected(ConfigurationException):
        config.configure()


class ConfigWithEntryPointClassList(Configuration):
    filename = 'config_file_for_this_egg.py'
    config_key = 'some_key'
    some_setting = EntryPointClassList('some.test.entrypoint', description='we test stuff')


class ListedClass1: pass
class ListedClass2: pass


@with_fixtures(ConfigWithFiles)
def test_entry_point_class_list(config_with_files):
    """EntryPointClassList is a special ConfigSetting which reads its value from a pkg_resources
       entry point which contains a list of classes published by any (possibly other) egg."""
    fixture = config_with_files
    
    # Because we cannot remove EasterEggs from pkg_resources, the next test must happen after
    # this one. The next check just ensures that we know when that does not happen:
    with expected(pkg_resources.DistributionNotFound):
        pkg_resources.require('test-inject')  

    fixture.set_config_spec(easter_egg, 'reahl.component_dev.test_config:ConfigWithEntryPointClassList')

    # Publish some classes on the entry point being tested
    line = 'ListedClass1 = reahl.component_dev.test_config:ListedClass1'
    easter_egg.add_entry_point_from_line('some.test.entrypoint', line)
    line = 'ListedClass2 = reahl.component_dev.test_config:ListedClass2'
    easter_egg.add_entry_point_from_line('some.test.entrypoint', line)

    # Usually this happens inside other infrastructure, such as the implementation of reahl serve (see reahl-dev)
    config = StoredConfiguration(fixture.config_dir.name)
    config.configure()

    # The classes are found from the entry point
    assert set(config.some_key.some_setting) == {ListedClass1, ListedClass2} 


class ConfigWithInjectedSetting(Configuration):
    filename = 'config_file_for_this_egg.py'
    config_key = 'some_key'
    injected_setting = ConfigSetting(automatic=True)


class ConfigWhichInjectsSetting(Configuration):
    filename = 'injector_egg.py'
    config_key = 'some_other_key'

    def do_injections(self, config):
        config.some_key.injected_setting = 123


@with_fixtures(ConfigWithFiles)
def test_config_defaults_automatic(config_with_files):
    """To facilitate dependency injection, the Configuration of one Egg can set the 'automatic' ConfigSettings of
       another egg on which it depends."""

    fixture = config_with_files
    
    egg_needing_injection = EasterEgg('test-inject')
    fixture.set_config_spec(egg_needing_injection, 'reahl.component_dev.test_config:ConfigWithInjectedSetting')
    pkg_resources.working_set.add(egg_needing_injection)

    fixture.set_config_spec(easter_egg, 'reahl.component_dev.test_config:ConfigWhichInjectsSetting')

    easter_egg.add_dependency(egg_needing_injection.as_requirement_string())

    # Usually this happens inside other infrastructure, such as the implementation of reahl serve (see reahl-dev)
    config = StoredConfiguration(fixture.config_dir.name)
    config.configure()

    # The default value is still used
    assert config.some_key.injected_setting == 123 


class ConfigWithDependentSetting(Configuration):
    filename = 'config_file_for_this_egg.py'
    config_key = 'some_key'
    some_setting = ConfigSetting(default='default value')
    some_other_setting = ConfigSetting(default=DeferredDefault(lambda c: 'tra %s lala' % c.some_setting))


@with_fixtures(ConfigWithFiles)
def test_config_dependent_defaults(config_with_files):
    """The default of one setting can be dependent on another setting if passed a callable"""
    
    fixture = config_with_files
    fixture.set_config_spec(easter_egg, 'reahl.component_dev.test_config:ConfigWithDependentSetting')

    # Usually this happens inside other infrastructure, such as the implementation of reahl serve (see reahl-dev)
    config = StoredConfiguration(fixture.config_dir.name)
    config.configure()

    # The setting was read
    assert config.some_key.some_setting == 'default value' 
    assert config.some_key.some_other_setting == 'tra default value lala' 

    



# Copyright 2013-2020 Reahl Software Services (Pty) Ltd. All rights reserved.
#
#    This file is part of Reahl.
#
#    Reahl is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation; version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.



import inspect
import warnings

import platform
import pkg_resources
import wrapt

def deprecated(message, version='n/a'):
    def catch_wrapped(f):
        def is_init_or_classmethod(member):
            if inspect.ismethod(member) and member.__self__ is f:
                return True
            return (inspect.ismethod(member) or inspect.isfunction(member)) and member.__name__ == '__init__'

        @wrapt.decorator
        def deprecated_wrapper(wrapped, instance, args, kwargs):
            if is_init_or_classmethod(wrapped):
                deprecated_thing = wrapped.__self__.__class__ 
                warnings.warn('DEPRECATED: %s. %s' % (deprecated_thing.__class__, message), DeprecationWarning, stacklevel=2)
            else:
                deprecated_thing = wrapped
                warnings.warn('DEPRECATED: %s. %s' % (deprecated_thing, message), DeprecationWarning, stacklevel=2)
                
            return wrapped(*args, **kwargs)

        if pkg_resources.parse_version(platform.python_version()) > pkg_resources.parse_version('3.6') and f.__doc__:
            f.__doc__ = '%s\n\n.. deprecated:: %s\n   %s' % (f.__doc__, version, message)

        if inspect.isclass(f):
            for name, method in inspect.getmembers(f, predicate=is_init_or_classmethod):
                setattr(f, name, deprecated_wrapper(method))
            return f
        else:
            return deprecated_wrapper(f)

    return catch_wrapped





# Copyright 2013-2022 Reahl Software Services (Pty) Ltd. All rights reserved.
#
#    This file is part of Reahl.
#
#    Reahl is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation; version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Utilities to manipulate underlying databases - sometimes via an ORM tool.

.. uml:: ../../../reahl-component/reahl/component_dev/database_components.puml

"""


import re
from contextlib import contextmanager
import urllib.parse

import pkg_resources 

from reahl.component.exceptions import ProgrammerError
from reahl.component.eggs import ReahlEgg
from reahl.component.migration import MigrationPlan


class CouldNotFindDatabaseControlException(Exception):
    def __init__(self, url):
        super().__init__(url)
        self.url = url


class SystemControl:
    """Used to control all aspects relating to the underlying database as per the configuration of the system.

    Any Reahl system is assumed to have a database backing it. Our
    dealings with a databases vary in two dimentions. The first of
    these dimantions is the different database backends that can be
    used. How one controls the backend itself (such as creating users,
    dropping databases, etc) differs depending on the particular
    backend used.

    Regardless of which backend is used, one can also use an ORM layer
    on top of the database. Things like instrumenting classes,
    creating database tables of dealing with database transactions are
    dependent on the ORM implementation used.

    The SystemControl allows you to control all these aspects of the
    database. It does so by delegating methods relating to the
    database backend to a :class:`DatabaseControl` of the correct
    kind, and by delegating the methods relating to ORM operations to
    an :class:`ORMControl` of the correct kind.

    The type of :class:`DatabaseControl` used is inferred from the
    configured `reahlsystem.connection_uri` in `reahl.config.py`. The
    actual :class:`ORMControl` used is configured as the
    `reahlsystem.orm_control` in `reahl.config.py`.

    A programmer should use the SystemControl of the system as
    programmatic interface to the database without regard for which
    technology-specific classes it delegates to in order to do its
    work.

    """
    def __init__(self, config):
        self.config = config
        self.connection_uri = self.config.reahlsystem.connection_uri
        self.db_control = self.db_control_for_uri(self.connection_uri)
        self.orm_control = self.config.reahlsystem.orm_control
        self.orm_control.system_control = self

    def connect(self, auto_commit=False):
        """Connects and logs into the database."""
        self.orm_control.connect(auto_commit=auto_commit)
        if self.db_control.is_in_memory:
            self.create_db_tables()

    @contextmanager
    def auto_connected(self):
        if self.connected:
            yield

        self.connect()
        try:
            yield
        finally:
            self.disconnect()

    @property
    def connected(self):
        """Returns True if the system is currently connected to the database."""
        return self.orm_control.connected

    def disconnect(self):
        """Disconnects from the database."""
        self.orm_control.disconnect()

    def managed_transaction(self):
        return self.orm_control.managed_transaction()

    def nested_transaction(self):
        return self.orm_control.nested_transaction()

    def finalise_session(self):
        self.orm_control.finalise_session()

    def commit(self):
        """Commits the database."""
        self.orm_control.commit()

    def rollback(self):
        """Rolls back the database."""
        self.orm_control.rollback()

    def db_control_for_uri(self, url):
        for i in self.config.reahlsystem.databasecontrols:
            if i.matches_uri(url):
                return i(url, self.config)
        raise CouldNotFindDatabaseControlException(url)

    def __getattr__(self, name):
        return getattr(self.db_control, name)

    def size_database(self):
        """Returns the current size of the database."""
        with self.orm_control.managed_transaction():
            return self.db_control.size_database(self.orm_control)

    def create_db_tables(self):
        """Creates the underlying database schema."""
        eggs_in_order = ReahlEgg.get_all_relevant_interfaces(self.config.reahlsystem.root_egg)
        with self.orm_control.managed_transaction() as transaction:
            return self.orm_control.create_db_tables(transaction, eggs_in_order)

    def drop_db_tables(self):
        """Removes the underlying database schema."""
        with self.orm_control.managed_transaction() as transaction:
            return self.orm_control.drop_db_tables(transaction)

    def initialise_database(self):
        """Ensures a new clean database exists, with a schema created. This drops an existing database if one is present."""
        self.drop_database()
        self.create_database()
        self.connect()
        try:
            self.create_db_tables()
        finally:
            self.disconnect()

    def migrate_db(self, explain=False):
        """Runs the database migrations relevant to the current system."""
        self.orm_control.migrate_db(ReahlEgg.interface_for(pkg_resources.get_distribution(self.config.reahlsystem.root_egg)), explain=explain)
        return 0

    def diff_db(self, output_sql=False):
        """Computes the changes in schema between the current database and what the current system expects."""
        return self.orm_control.diff_db(output_sql=output_sql)

    def do_daily_maintenance(self):
        """Runs the all the scheduled jobs relevant to the current system."""
        with self.orm_control.managed_transaction() as transaction:
            ReahlEgg.do_daily_maintenance_for_egg(self.config.reahlsystem.root_egg)

    def create_db_user(self, super_user_name=None, create_with_password=True):
        """Creates the database user."""
        return self.db_control.create_db_user(super_user_name=super_user_name, create_with_password=create_with_password)

    def drop_db_user(self, super_user_name=None):
        """Drops the database user."""
        return self.db_control.drop_db_user(super_user_name=super_user_name)
    
    def drop_database(self, super_user_name=None):
        """Drops the database (if it exists)."""
        return self.db_control.drop_database(super_user_name=super_user_name)
    
    def create_database(self, super_user_name=None):
        """Creates the database."""
        return self.db_control.create_database(super_user_name=super_user_name)
    
    def backup_database(self, directory, super_user_name=None):
        """Backs up the database."""
        return self.db_control.backup_database(directory, super_user_name=super_user_name)
    
    def backup_all_databases(self, directory, super_user_name=None):
        """Backs up all databases on the current machine."""
        return self.db_control.backup_all_databases(directory, super_user_name=super_user_name)
    
    def restore_database(self, filename, super_user_name=None):
        """Restores a databases from the given backup."""
        return self.db_control.restore_database(filename, super_user_name=super_user_name)
    
    def restore_all_databases(self, filename, super_user_name=None):
        """Restores all databases from the given backup."""
        return self.db_control.restore_all_databases(filename, super_user_name=super_user_name)

    

class DatabaseControl:
    """An interface to the underlying database backend technology used. 

    This class has the responsibility to manage the low-level backend
    operations, such as creating/dropping databases or users.

    In order to support a new type of database, subclass this class
    and implement its methods appropriately for that backend.

    Don't use this class directly, rather call the methods on the
    :class:`SystemControl`. It delegates to the appropriate underlying
    :class:`ORMControl` and/or :class:`DatabaseControl` as
    appropriate.
    """
    control_matching_regex = r'^$'
    def __init__(self, url, config):
        self.config = config
        self.connection_uri = url
        uri_parts = urllib.parse.urlparse(url)
        self.user_name = self.unquote(uri_parts.username)
        self.password = self.unquote(uri_parts.password)
        self.host = self.unquote(uri_parts.hostname)
        self.port = uri_parts.port
        self.database_name = self.unquote(uri_parts.path[1:] if uri_parts.path.startswith('/') else uri_parts.path)
        if not self.database_name:
            raise ProgrammerError('Please specify a database name in reahlsystem.connection_uri')

    def unquote(self, value):
        return urllib.parse.unquote(value) if value else None

    def get_dbapi_connection_creator(self):
        return None

    @property
    def is_in_memory(self):
        return False

    @classmethod
    def matches_uri(cls, url):
        return re.match(cls.control_matching_regex, url) is not None

    def create_db_user(self, super_user_name=None, create_with_password=True):
        assert self.user_name, 'No user name in URI %s' % self.connection_uri

    def drop_db_user(self, super_user_name=None):
        assert self.user_name, 'No user name in URI %s' % self.connection_uri

    
class NullDatabaseControl(DatabaseControl):
    """A stubbed-out :class:`DatabaseControl` for systems that do not have any database at all."""

    def donothing(self, *args, **kwargs):
        pass

    def __getattr__(self, name):
        return self.donothing


class ORMControl:
    """An interface to higher-level database operations that may be dependent on the ORM technology used.

    This class has the responsibility to manage the higher-level
    backend operations, such as instrumenting persisted classes,
    creating database schemas, migration, etc.

    In order to support a new type of ORM, subclass this class
    and implement its methods appropriately for that ORM.

    Don't use this class directly, rather call the methods on the
    :class:`SystemControl`. It delegates to the appropriate underlying
    :class:`ORMControl` and/or :class:`DatabaseControl` as
    appropriate.

    .. versionchanged: 5.0
       Signature changed from taking eggs_in_order to taking root_egg.
    """
    def migrate_db(self, root_egg, explain=False):
        plan = MigrationPlan(root_egg, self)
        if explain:
            try:
                plan.do_planning()
            finally:
                plan.explain()
        else:
            plan.do_planning()
            plan.execute()
                   

    @contextmanager
    def nested_transaction(self):
        """A context manager for code that needs to run in a nested transaction.

        .. versionchanged:: 5.0
           Changed to yield a TransactionVeto which can be used to override when the transaction will be committed or not.
        """
        pass

    @contextmanager
    def managed_transaction(self):
        pass
    def connect(self, auto_commit=False):
        pass
    @property
    def connected(self):
        pass
    def finalise_session(self):
        pass
    def disconnect(self):
        pass
    def commit(self):
        """Commits the current transaction. Programmers should not need to deal with such transaction
           management explicitly, since the framework already manages transactions itself."""
        pass
    def rollback(self):
        """Rolls back the current transaction. Programmers should not need to deal with such transaction
           management explicitly, since the framework already manages transactions itself."""
        pass
    def create_db_tables(self, transaction, eggs_in_order):
        pass
    def drop_db_tables(self, transaction):
        pass
    def execute_one(self, sql):
        pass
    def prune_schemas_to_only(self, live_versions):
        pass
    def diff_db(self, output_sql=False):
        pass
    def initialise_schema_version_for(self, egg=None, egg_name=None, egg_version=None):
        pass
    def remove_schema_version_for(self, egg=None, egg_name=None, fail_if_not_found=True):
        pass
    def schema_version_for(self, egg, default=None):
        pass
    def set_schema_version_for(self, version):
        pass






# Copyright 2013-2022 Reahl Software Services (Pty) Ltd. All rights reserved.
#
#    This file is part of Reahl.
#
#    Reahl is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation; version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
from contextlib import contextmanager
import warnings
import re

from reahl.tofu import Fixture, expected, temp_dir, uses, scenario
from reahl.tofu.pytestsupport import with_fixtures
from reahl.stubble import CallMonitor, EmptyStub, stubclass

from reahl.component.dbutils import ORMControl
from reahl.component.eggs import ReahlEgg, Version, Dependency, CircularDependencyDetected, InvalidDependencySpecification
from reahl.component.migration import Migration, MigrationPlan, MigrationSchedule, ExceptionDuringMigration
from reahl.component.exceptions import ProgrammerError


@stubclass(Dependency)
class StubDependency:
    type = 'egg'
    distribution = EmptyStub()
    def __init__(self, version, egg_type='egg', is_component=True):
        self.version = version
        self.name = self.version.name
        self.type = egg_type
        self._is_component = is_component

    def get_best_version(self):
        return self.version

    def __str__(self):
        return str(self.version)

    @property
    def is_component(self):
        return self._is_component


@stubclass(ReahlEgg)
class ReahlEggStub(ReahlEgg):
    def __init__(self, name, version_info):
        super().__init__(None)
        self._name = name
        self.version_info = version_info
        self.dependencies = {}

    def create_metadata(self, distribution):
        return {'metadata_version': str(ReahlEgg.metadata_version)}
        
    @property
    def name(self):
        return self._name

    def get_versions(self):
        unsorted_versions = [Version(self, version_number_string) for version_number_string in self.version_info]
        return list(sorted(unsorted_versions, key=lambda x: x.version_number))

    @property
    def installed_version(self):
        return self.get_versions()[-1]

    def get_migration_classes_for_version(self, version):
        return self.version_info[str(version.version_number)]

    def get_dependencies(self, version):
        return self.dependencies.get(version, [])




@stubclass(ORMControl)
class ORMControlStub(ORMControl):
    created_schema_for = None
    pruned_schemas_to = []

    def __init__(self):
        self.versions = {}

    @contextmanager
    def managed_transaction(self):
        yield

    def set_schema_version_for(self, version):
        self.versions[version.egg.name] = str(version.version_number)

    def schema_version_for(self, egg, default=None):
        return self.versions.get(egg.name, '0.0')

    def initialise_schema_version_for(self, egg=None, egg_name=None, egg_version=None):
        self.versions[egg.name] = str(egg.version.version_number)

    def prune_schemas_to_only(self, live_versions):
        self.pruned_schemas_to = live_versions



class MigrateFixture(Fixture):
    def new_orm_control(self):
        return ORMControlStub()

    def new_some_object(self):
        class SomeObject:
            calls_made = []

            def do_something(self, arg):
                self.calls_made.append(arg)

        return SomeObject()

    def check_order(self, operations):
        keys = list(set([i.split('-')[-1] for i in operations]))
        for key in keys:
            try:
                drop_pk_index = operations.index('drop_pk-%s' % key)
            except ValueError:
                drop_pk_index = -1
            try:
                create_pk_index = operations.index('create_pk-%s' % key)
            except ValueError:
                create_pk_index = -1

            create_fk_indexes = [i for i in range(len(operations)) if operations[i] == 'create_fk-%s' % key]
            drop_fk_indexes = [i for i in range(len(operations)) if operations[i] == 'drop_fk-%s' % key]

            assert all([i > create_pk_index for i in create_fk_indexes])   # creating a pk happens before creating a fk to it
            assert drop_pk_index < 0 or drop_pk_index > create_pk_index    # dropping a pk happens after creating it

            for i in range(len(create_fk_indexes)):                        # dropping an fk happens after creating it (multiple times)
                if len(drop_fk_indexes) > i:
                    assert create_fk_indexes[i] < drop_fk_indexes[i]

        
@with_fixtures(MigrateFixture)
def test_how_migration_works(fixture):
    """A logical change to the database is coded in a Migration. In a Migration, override
       schedule_upgrades in which changes are scheduled to be run during the appropriate phase.
    """

    some_object = fixture.some_object
    
    class Migration1(Migration):

        def schedule_upgrades(self):
            self.schedule('drop_fk', some_object.do_something, 'drop_fk-1')
            self.schedule('data', some_object.do_something, 'data-1')
            self.schedule('drop_fk', some_object.do_something, 'drop_fk-2')

    class Migration2(Migration):

        def schedule_upgrades(self):
            self.schedule('drop_fk', some_object.do_something, 'drop_fk-3')

    egg = ReahlEggStub('my_egg', {'1.0': [], '1.1': [Migration1, Migration2]})
    fixture.orm_control.set_schema_version_for(egg.get_versions()[0])

    fixture.orm_control.migrate_db(egg)

    expected_order = ['drop_fk-1', 'drop_fk-2', 'drop_fk-3', 'data-1']
    assert some_object.calls_made == expected_order


@with_fixtures(MigrateFixture)
def test_migrating_dependencies(fixture):
    """Only the necessary Migrations are run to bring the database schema up to date from a previous running installation.
    """

    some_object = fixture.some_object
    
    class MainMigration1(Migration):
        def schedule_upgrades(self):
            self.schedule('create_fk', some_object.do_something, 'create_fk-1')

    class MainMigration2(Migration):
        def schedule_upgrades(self):
            self.schedule('drop_fk', some_object.do_something, 'drop_fk-1')
            self.schedule('create_fk', some_object.do_something, 'create_fk-2')


    orm_control = fixture.orm_control
    main_egg = ReahlEggStub('main_egg', {'1.0': [MainMigration1], '1.1': [MainMigration2]})


    class DependencyMigration1(Migration):
        def schedule_upgrades(self):
            self.schedule('create_pk', some_object.do_something, 'create_pk-1')

    class DependencyMigration2(Migration):
        def schedule_upgrades(self):
            self.schedule('drop_pk', some_object.do_something, 'drop_pk-1')
            self.schedule('create_pk', some_object.do_something, 'create_pk-2')

    dependency_egg = ReahlEggStub('dependency_egg', {'5.0': [DependencyMigration1], '5.1': [DependencyMigration2]})

    [mv1, mv2] = main_egg.get_versions()
    [dv1, dv2] = dependency_egg.get_versions()

    main_egg.dependencies = {str(mv1.version_number): [StubDependency(dv1)],
                             str(mv2.version_number): [StubDependency(dv2)]}

    fixture.orm_control.migrate_db(main_egg)

    expected_order = ['create_pk-1', 'create_fk-1', 'drop_fk-1', 'drop_pk-1', 'create_pk-2', 'create_fk-2']
    assert some_object.calls_made == expected_order


@uses(migrate_fixture=MigrateFixture)
class DependencyScenarios(Fixture):

    @property
    def orm_control(self):
        return self.migrate_fixture.orm_control

    @scenario
    def egg_dependency(self):
        self.egg_type = 'egg'
        self.is_component = True

    @scenario
    def thirdparty_component_dependency(self):
        self.egg_type = 'thirdparty'
        self.is_component = True

    @scenario
    def thirdparty_dependency(self):
        self.egg_type = 'thirdparty'
        self.is_component = False


@with_fixtures(DependencyScenarios)
def test_dependency_types_detected(dependency_scenarios):
    """Components may depend on other components. These dependencies may be referred to as egg or thirdparty dependencies.
    If these dependencies are components, they should be included in the version dependency graph.
    """

    main_egg = ReahlEggStub('main_egg', {'1.0': []})

    dependency_egg = ReahlEggStub('dependency_egg', {'5.0': []})

    [mv1] = main_egg.get_versions()
    [dv1] = dependency_egg.get_versions()

    main_egg.dependencies = {str(mv1.version_number): 
                             [StubDependency(dv1, egg_type=dependency_scenarios.egg_type,
                                                  is_component=dependency_scenarios.is_component)]}

    plan = MigrationPlan(main_egg, dependency_scenarios.orm_control)
    plan.do_planning()

    if dependency_scenarios.is_component:
        assert dv1 in plan.version_graph.graph
    else:
        assert dv1 not in plan.version_graph.graph


@with_fixtures(MigrateFixture)
def test_migrating_dependencies_with_intermediate_versions(fixture):
    """A dependency on a project can skip intermediate versions of the project, yet the necessary migrations are still run.
    """

    some_object = fixture.some_object
    
    class MainMigration1(Migration):
        def schedule_upgrades(self):
            self.schedule('create_fk', some_object.do_something, 'MainMigration1 create_fk-1')

    class MainMigration2(Migration):
        def schedule_upgrades(self):
            self.schedule('drop_fk', some_object.do_something, 'MainMigration2 drop_fk-1')
            self.schedule('create_fk', some_object.do_something, 'MainMigration2 create_fk-4')

    orm_control = fixture.orm_control
    main_egg = ReahlEggStub('main_egg', {'1.0': [MainMigration1], '1.1': [MainMigration2]})

    class DependencyMigration11(Migration):
        def schedule_upgrades(self):
            self.schedule('create_fk', some_object.do_something, 'DependencyMigration11 create_fk-1')

    class DependencyMigration12(Migration):
        def schedule_upgrades(self):
            self.schedule('drop_fk', some_object.do_something, 'DependencyMigration12 drop_fk-1')
            self.schedule('create_fk', some_object.do_something, 'DependencyMigration12 create_fk-2')

    class DependencyMigration13(Migration):
        def schedule_upgrades(self):
            self.schedule('drop_fk', some_object.do_something, 'DependencyMigration13 drop_fk-2')
            self.schedule('create_fk', some_object.do_something, 'DependencyMigration13 create_fk-4')

    dependency_egg1 = ReahlEggStub('dependency_egg1', {'1.1': [DependencyMigration11], '1.2': [DependencyMigration12], '1.3': [DependencyMigration13]})

    class DependencyMigration21(Migration):
        def schedule_upgrades(self):
            self.schedule('create_pk', some_object.do_something, 'DependencyMigration21 create_pk-1')

    class DependencyMigration22(Migration):
        def schedule_upgrades(self):
            self.schedule('drop_pk', some_object.do_something, 'DependencyMigration22 drop_pk-1')
            self.schedule('create_pk', some_object.do_something, 'DependencyMigration22 create_pk-2')

    class DependencyMigration23(Migration):
        def schedule_upgrades(self):
            self.schedule('drop_pk', some_object.do_something, 'DependencyMigration23 drop_pk-2')
            self.schedule('create_pk', some_object.do_something, 'DependencyMigration23 create_pk-3')

    class DependencyMigration24(Migration):
        def schedule_upgrades(self):
            self.schedule('drop_pk', some_object.do_something, 'DependencyMigration24 drop_pk-3')
            self.schedule('create_pk', some_object.do_something, 'DependencyMigration24 create_pk-4')

    dependency_egg2 = ReahlEggStub('dependency_egg2', {'2.1': [DependencyMigration21], '2.2': [DependencyMigration22], '2.3': [DependencyMigration23], '2.4': [DependencyMigration24]})

    [v1, v2] = main_egg.get_versions()
    [dv11, dv12, dv13] = dependency_egg1.get_versions()
    [dv21, dv22, dv23, dv24] = dependency_egg2.get_versions()

    main_egg.dependencies = {str(v1.version_number): [StubDependency(dv11), StubDependency(dv21)],
                             str(v2.version_number): [StubDependency(dv13), StubDependency(dv24)]}  # Jump over dv12

    dependency_egg1.dependencies = {str(dv11.version_number): [StubDependency(dv21)],
                                    str(dv12.version_number): [StubDependency(dv22)],
                                    str(dv13.version_number): [StubDependency(dv24)]}  # Jump over dv23

    fixture.orm_control.migrate_db(main_egg)

    expected_order = [
         'create_pk-1',
         'create_fk-1', 
         'create_fk-1',
         'drop_fk-1', 
         'drop_fk-1',
         
         'drop_pk-1',
         
         'create_pk-2',
         
         'create_fk-2',
         
         'drop_fk-2',
         
         'drop_pk-2',
         
         'create_pk-3',
         'drop_pk-3',
         'create_pk-4',
         'create_fk-4',
         'create_fk-4']

    assert [i.split()[1] for i in some_object.calls_made] == expected_order


@with_fixtures(MigrateFixture)
def test_migrating_from_existing_schema(fixture):
    """The Migrations of all dependencies and all previous versions (and their dependencies) are
       scheduled.
    """

    some_object = fixture.some_object
    
    class MainMigration1(Migration):
        def schedule_upgrades(self):
            self.schedule('create_fk', some_object.do_something, 'create_fk_1')

    class MainMigration2(Migration):
        def schedule_upgrades(self):
            self.schedule('drop_fk', some_object.do_something, 'drop_fk_1')
            self.schedule('create_fk', some_object.do_something, 'create_fk_2')


    orm_control = fixture.orm_control
    main_egg = ReahlEggStub('main_egg', {'1.0': [MainMigration1], '1.1': [MainMigration2]})


    class DependencyMigration1(Migration):
        def schedule_upgrades(self):
            self.schedule('create_pk', some_object.do_something, 'create_pk_1')

    class DependencyMigration2(Migration):
        def schedule_upgrades(self):
            self.schedule('drop_pk', some_object.do_something, 'drop_pk_1')
            self.schedule('create_pk', some_object.do_something, 'create_pk_2')

    dependency_egg = ReahlEggStub('dependency_egg', {'5.0': [DependencyMigration1], '5.1': [DependencyMigration2]})

    [mv1, mv2] = main_egg.get_versions()
    [dv1, dv2] = dependency_egg.get_versions()

    main_egg.dependencies = {str(mv1.version_number): [StubDependency(dv1)],
                             str(mv2.version_number): [StubDependency(dv2)]}

    fixture.orm_control.set_schema_version_for(mv1)
    fixture.orm_control.set_schema_version_for(dv1)
    fixture.orm_control.migrate_db(main_egg)

    expected_order = ['drop_fk_1', 'drop_pk_1', 'create_pk_2', 'create_fk_2']
    assert some_object.calls_made == expected_order


@with_fixtures(MigrateFixture)
def test_migrating_changing_dependencies(fixture):
    """The dependencies of Versions can change during the evolution of the relateOnly the necessary Migrations are run
       to bring the database schema up to date from a previous running installation.
    """

    some_object = fixture.some_object
    
    orm_control = fixture.orm_control

    class MainMigration1(Migration):
        def schedule_upgrades(self):
            self.schedule('create_fk', some_object.do_something, '1 create_fk-21')
            self.schedule('create_fk', some_object.do_something, '1 create_fk-11')

    class MainMigration2(Migration):
        def schedule_upgrades(self):
            self.schedule('drop_fk', some_object.do_something, '2 drop_fk-21')
            self.schedule('drop_fk', some_object.do_something, '2 drop_fk-11')
            self.schedule('create_fk', some_object.do_something, '2 create_fk-12')

    main_egg = ReahlEggStub('main_egg', {'1.0': [MainMigration1], '1.1': [MainMigration2]})

    class DependencyMigration11(Migration):
        def schedule_upgrades(self):
            self.schedule('create_fk', some_object.do_something, '11 create_fk-21')
            self.schedule('create_pk', some_object.do_something, '11 create_pk-11')

    class DependencyMigration12(Migration):
        def schedule_upgrades(self):
            self.schedule('drop_fk', some_object.do_something, '12 drop_fk-21')
            self.schedule('drop_pk', some_object.do_something, '12 drop_pk-11')
            self.schedule('create_pk', some_object.do_something, '12 create_pk-12')

    dependency_egg1 = ReahlEggStub('dependency_egg1', {'5.0': [DependencyMigration11], '5.1': [DependencyMigration12]})

    class DependencyMigration21(Migration):
        def schedule_upgrades(self):
            self.schedule('create_pk', some_object.do_something, '21 create_pk-21')

    class DependencyMigration22(Migration):
        def schedule_upgrades(self):
            self.schedule('create_fk', some_object.do_something, '22 create_fk-12')

    dependency_egg2 = ReahlEggStub('dependency_egg2', {'3.2': [DependencyMigration21], '3.3': [DependencyMigration22]})

    [mv1, mv2] = main_egg.get_versions()
    [dv11, dv12] = dependency_egg1.get_versions()
    [dv21, dv22] = dependency_egg2.get_versions()

    main_egg.dependencies = {str(mv1.version_number): [StubDependency(dv11), StubDependency(dv21)],
                             str(mv2.version_number): [StubDependency(dv12)]} # and no dependency on dependency_egg2, thus no migrations expected

    dependency_egg1.dependencies = {str(dv11.version_number): [StubDependency(dv21)]}
    dependency_egg2.dependencies = {str(dv22.version_number): [StubDependency(dv12)]}  # Dependency gets swapped


    fixture.orm_control.migrate_db(main_egg)

    fixture.check_order([i.split()[1] for i in some_object.calls_made])

    assert '22 create_fk-12' not in some_object.calls_made # Migrations of previously used, but now-dead components are not run
    assert dv22 not in fixture.orm_control.pruned_schemas_to # dv22 is not live anymore, and its schema is cleaned up
    assert all([i in fixture.orm_control.pruned_schemas_to for i in [mv2, dv12]]) # These schemas are still live


@with_fixtures(MigrateFixture)
def test_intra_cluster_circular_dependency(fixture):
    """Circular dependencies within a cluster are not allowed.
    """

    some_object = fixture.some_object

    orm_control = fixture.orm_control
    main_egg = ReahlEggStub('main_egg', {'1.0': []})
    dependency_egg = ReahlEggStub('dependency_egg', {'5.0': []})

    [mv1] = main_egg.get_versions()
    [dv1] = dependency_egg.get_versions()

    main_egg.dependencies = {str(mv1.version_number): [StubDependency(dv1)]}
    dependency_egg.dependencies = {str(dv1.version_number): [StubDependency(mv1)]}
    
    with expected(CircularDependencyDetected, test=r'(dependency_egg\[5\.0\] -> main_egg\[1\.0\] -> dependency_egg\[5\.0\]|main_egg\[1\.0\] -> dependency_egg\[5\.0\] -> main_egg\[1\.0\])'):
        fixture.orm_control.migrate_db(main_egg)


@with_fixtures(MigrateFixture)
def test_dependencies_resulting_in_duplicates_in_cluster(fixture):
    """Dependencies resulting in installing more than one version of the same package are not allowed.
    """

    some_object = fixture.some_object

    orm_control = fixture.orm_control
    main_egg = ReahlEggStub('main_egg', {'1.0': [], '1.1': []})
    dependency_egg = ReahlEggStub('dependency_egg', {'5.0': [], '5.1': []})

    [mv1, mv2] = main_egg.get_versions()
    [dv1, dv2] = dependency_egg.get_versions()

    main_egg.dependencies = {str(mv1.version_number): [StubDependency(dv1)],
                             str(mv2.version_number): [StubDependency(dv2)]}
    dependency_egg.dependencies = {str(dv2.version_number): [StubDependency(mv1)]}

    with expected(InvalidDependencySpecification, test=r'Dependencies result in installing more than one version of: dependency_egg,main_egg\. Dependencies: main_egg\[1\.0\] -> \[dependency_egg\[5\.0\]\],dependency_egg\[5\.1\] -> \[main_egg\[1.0\]\],main_egg\[1\.1\] -> \[dependency_egg\[5\.1\]\]'):
        fixture.orm_control.migrate_db(main_egg)


def test_schedule_executes_in_order():
    """A MigrationSchedule is used internally to schedule calls in different phases. The calls 
       scheduled in each phase are executed in the order the phases have been set up on the MigrationSchedule.
       Within a phase, the calls are executed in the order they were registered in that phase.
    """
    
    migration_schedule = MigrationSchedule(EmptyStub(), EmptyStub(), [])

    class SomeObject:
        def do_something(self, arg):
            pass
    some_object = SomeObject()

    # schedule calls not in registered order
    with CallMonitor(some_object.do_something) as monitor:
        migration_schedule.schedule('cleanup', EmptyStub(), EmptyStub(), some_object.do_something, '1')
        migration_schedule.schedule('create_fk', EmptyStub(), EmptyStub(), some_object.do_something, '2')
        migration_schedule.schedule('data', EmptyStub(), EmptyStub(), some_object.do_something, '3')
        migration_schedule.schedule('indexes', EmptyStub(), EmptyStub(), some_object.do_something, '4')
        migration_schedule.schedule('create_pk', EmptyStub(), EmptyStub(), some_object.do_something, '5')

        migration_schedule.schedule('alter', EmptyStub(), EmptyStub(), some_object.do_something, 'c1')
        migration_schedule.schedule('drop_pk', EmptyStub(), EmptyStub(),some_object.do_something, 'a1')
        migration_schedule.schedule('pre_alter', EmptyStub(), EmptyStub(),some_object.do_something, 'b')
        migration_schedule.schedule('drop_pk', EmptyStub(), EmptyStub(),some_object.do_something, 'a2')
        migration_schedule.schedule('alter', EmptyStub(), EmptyStub(),some_object.do_something, 'c2')

    migration_schedule.execute_all()

    actual_order = [call.args[0] for call in monitor.calls]
    expected_order = ['a1', 'a2', 'b', 'c1', 'c2', '5', '4', '3', '2', '1']
    assert actual_order == expected_order


def test_schedule_executes_phases_with_parameters():
    """When a MigrationSchedule executes the calls that were scheduled from a Migration, 
       the methods are actually called, and passed the correct arguments."""

    class SomeObject:
        def please_call_me(self, arg, kwarg=None):
            pass
    some_object = SomeObject()
    
    migration_schedule = MigrationSchedule(EmptyStub(), EmptyStub(), [])
    migration = Migration(migration_schedule)

    with CallMonitor(some_object.please_call_me) as monitor:
        migration.schedule('alter', some_object.please_call_me, 'myarg', kwarg='mykwarg')

    migration_schedule.execute_all()

    assert monitor.calls[0].args == ('myarg',)
    assert monitor.calls[0].kwargs == dict(kwarg='mykwarg')


def test_invalid_schedule_name_raises():
    """A useful error is raised when an attempt is made to schedule a call in a phase that is not defined."""
    
    migration_schedule = MigrationSchedule(EmptyStub(), EmptyStub(), [])

    with expected(ProgrammerError, test=r'A phase with name<wrong_name> does not exist\.'):
        migration_schedule.schedule('wrong_name', EmptyStub(), EmptyStub(), EmptyStub())


def test_missing_schedule_upgrades_warns():
    """If a programmer does not override schedule_upgrades, a warning is raised."""
    class TestMigration(Migration):
        pass

    with warnings.catch_warnings(record=True) as raised_warnings:
        warnings.simplefilter("always")
  
        TestMigration(EmptyStub()).schedule_upgrades()
 
    [warning] = raised_warnings
    expected_message = 'Ignoring TestMigration.schedule_upgrades(): it does not override schedule_upgrades() ' \
                       '(method name typo perhaps?)'
    assert str(warning.message) == expected_message


def test_error_reporting_on_breaking_migrations():
    """When there is an error during execution of a Migration, the code where it was scheduled is reported."""

    class SomeObject:
        def please_call_me(self):
            raise Exception('breaking intentionally')
    some_object = SomeObject()

    migration_schedule = MigrationSchedule(EmptyStub(), EmptyStub(), [])
    migration = Migration(migration_schedule)

    migration.schedule('alter', some_object.please_call_me)

    def is_please_call_me_on_stack(ex):
        return re.match(".*The above Exception happened for the migration that was scheduled here:.*    migration.schedule\('alter', some_object.please_call_me\)", str(ex), re.MULTILINE|re.S)
    with expected(ExceptionDuringMigration, test=is_please_call_me_on_stack):
        migration_schedule.execute_all()


@with_fixtures(MigrateFixture)
def test_planning(fixture):
    """A plan can be explained by generating graphs used by the migration algorithm.
    """

    egg = ReahlEggStub('my_egg', {'1.0': [], '1.1': []})

    plan = MigrationPlan(egg, fixture.orm_control)
    plan.do_planning()

    with temp_dir().as_cwd() as dir_name:
        plan.explain()
        assert os.path.isfile(os.path.join(dir_name, 'clusters.svg'))
        assert os.path.isfile(os.path.join(dir_name, 'versions.svg'))
        assert os.path.isfile(os.path.join(dir_name, 'schedules.svg'))






# Copyright 2013-2022 Reahl Software Services (Pty) Ltd. All rights reserved.
#
#    This file is part of Reahl.
#
#    Reahl is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation; version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Exceptions used throughout several Reahl components."""

import sys

import wrapt
from wrapt.wrappers import PartialCallableObjectProxy
import inspect


from collections.abc import Callable

from reahl.component.context import ExecutionContext
from reahl.component.context import ExecutionContext, NoContextFound


class DomainException(Exception):
    """Any exception indicating an application-specific error condition that 
       should be communicated to a user.

       :keyword commit: Set to True to indicate that the current database transaction 
                        should be committed. By default transactions are rolled back
                        when a DomainException is raised.
       :keyword message: Optional error message.
       :keyword detail_messages: A list of error messages giving more detail about the exception.
       :keyword handled_inline: If False, indicates that this exception is not reported in the normal rendering of the page.

       .. versionchanged: 5.0
          Added `detail_messages` kwarg.
    """
    def __init__(self, commit=False, message=None, detail_messages=[], handled_inline=True, json_string=None):
        super().__init__(message)
        self.commit = commit
        self.message = message
        self.detail_messages = detail_messages
        self.handled_inline = handled_inline
        self.json_string = json_string
    
    def __reduce__(self):
        return (self.__class__, (self.commit, self.message, self.detail_messages, self.handled_inline, self.json_string))

    def as_user_message(self):
        # To ensure this module can be imported at a very low level
        from reahl.component.i18n import Catalogue
        _ = Catalogue('reahl-component')
        return self.message if self.message else _('An error occurred: %s' % self.__class__.__name__)

    def as_json(self):
        if self.json_string:
            return self.json_string
        else:
            return '"%s"' % super().__str__()

class AccessRestricted(Exception):
    """Raised to prevent the current user to perform some function which is not allowed."""

class ProgrammerError(Exception):
    """Raised to indicate an error in the program."""

class IncorrectArgumentError(ProgrammerError):
    """Raised to indicate an attempt to pass an incorrect argument to a Python callable."""
    def __init__(self, explanation, cause):
        ProgrammerError.__init__(self)
        self.explanation = explanation
        self.cause = cause
    
    def __str__(self):
        return '%s (%s)' % (self.explanation, self.cause)


class NotYetAvailable:
    def __init__(self, name):
        self.name = name
    def __str__(self):
        return '<%s name=%s>' % (self.__class__.__name__, self.name)

class DeferredImport:
    def __init__(self, value_or_string):
        self.value_or_string = value_or_string

    @property
    def value(self):
        return self.coerce_to_type(self.value_or_string)

    def import_string_spec(self, string_spec):
        bits = string_spec.split(':')
        assert len(bits) == 2, 'Invalid specification %s' % string_spec
        [module_name, class_name] = bits

        try:
            module = __import__(module_name, {}, {}, [class_name], 0)
            return getattr(module, class_name)
        except ImportError:
            raise ProgrammerError('Could not import %s' % string_spec)
        except AttributeError:
            raise ProgrammerError('Could not find %s in %s (from %s)' % (class_name, module_name, string_spec))

    def coerce_to_type(self, type_or_string):
        if isinstance(type_or_string, str):
            return self.import_string_spec(type_or_string)
        else:
            return type_or_string

class ArgumentCheck(Exception):
    def __init__(self, allow_none=False):
        self.allow_none = allow_none

    def check(self, func, name, value):
        if isinstance(value, NotYetAvailable):
            if value.name != name:
                raise IncorrectArgumentError('expected an argument for %s, got ' % name, value)
            else:
                return

        if value is None:
            if not self.allow_none:
                self.raise_with(func, name, value)
            else:
                return
        if not self.is_valid(value):
            self.raise_with(func, name, value)

    def is_valid(self, value):
        return False
        
    def raise_with(self, func, arg_name, value):
        self.func = func
        self.arg_name = arg_name
        self.value = value
        raise self

class TypeBasedArgumentCheck(ArgumentCheck):
    def __init__(self, type_or_string, allow_none=False):
        super().__init__(allow_none=allow_none)
        self.type_ = DeferredImport(type_or_string)

    def is_marked_with_attribute(self, value):
        return hasattr(self.type_.value, '__name__') and hasattr(value, 'is_%s' % self.type_.value.__name__)

class IsInstance(TypeBasedArgumentCheck):
    def is_valid(self, value):
        return isinstance(value, self.type_.value) or self.is_marked_with_attribute(value)

    def __str__(self):
        return '%s: %s should be an instance of %s (got %s instead)' % (self.func, self.arg_name, self.type_.value, self.value)

class IsSubclass(TypeBasedArgumentCheck):
    def is_valid(self, value):
        return inspect.isclass(value) and (issubclass(value, self.type_.value) or self.is_marked_with_attribute(value))

    def __str__(self):
        return '%s: %s should be %s or subclass of it (got %s instead)' % (self.func, self.arg_name, self.type_.value, self.value)


class IsCallable(ArgumentCheck):
    def __init__(self, args=(), kwargs={}, allow_none=False):
        super().__init__(allow_none=allow_none)
        self.args = args
        self.kwargs = kwargs

    def check(self, func, name, value):
        super().check(func, name, value)
        if value:
            message = '%s will be called with %s' % (value, self.formatted_message())
            ArgumentCheckedCallable(value, explanation=message).checkargs(*self.args, **self.kwargs)

    def formatted_message(self):
        formatted_args = ','.join([i.name for i in self.args])
        formatted_kwargs = ','.join(['%s=%s' for name, default in self.kwargs.items()])
        formatted_all = []
        if formatted_args:
            formatted_all.append(formatted_args)
        if formatted_kwargs:
            formatted_all.append(formatted_kwargs)
        return '(%s)' % (','.join(formatted_all))

    def is_valid(self, value):
        return isinstance(value, Callable)

    def __str__(self):
        return '%s: %s should be a callable object (got %s)' % (self.func, self.arg_name, self.value)

class ArgumentCheckedCallable:
    def __init__(self, target, explanation=None):
        self.target = target
        self.explanation = explanation

    def __call__(self, *args, **kwargs):
        self.checkargs(*args, **kwargs)
        return self.target(*args, **kwargs)

    def checkargs(self, *args, **kwargs):

        try:
            config = ExecutionContext.get_context().config
            if not config.reahlsystem.runtime_checking_enabled:
                return
        except (NoContextFound, AttributeError):
            pass

        if isinstance(self.target, PartialCallableObjectProxy):
            to_check = self.target.__call__
        elif inspect.ismethod(self.target):
            to_check = self.target
        elif inspect.isfunction(self.target):
            to_check = self.target
        elif inspect.isclass(self.target):
            to_check = self.target.__init__
            args = (NotYetAvailable('self'), )+args
        elif isinstance(self.target, Callable):
            to_check = self.target.__call__
        else:
            raise ProgrammerError('%s was expected to be a callable object' % self.target)

        try:
            try:
                bound_args = inspect.getcallargs(to_check, *args, **kwargs)
            except TypeError as ex:
                ex.args = (('%s: ' % self.target)+ex.args[0],) + ex.args[1:]
                raise
            args_to_check = getattr(to_check, 'arg_checks', {})
            for arg_name, arg_check in args_to_check.items():
                if arg_name in bound_args.keys():
                    arg_check.check(self.target, arg_name, bound_args[arg_name])
        except (TypeError, ArgumentCheck) as ex:
            if self.explanation:
                _, _, tb = sys.exc_info()
                new_ex = IncorrectArgumentError(self.explanation, ex)
                raise new_ex.with_traceback(tb)
            else:
                raise


def arg_checks(**checks):
    def catch_wrapped(f):
        if inspect.ismethoddescriptor(f):
            f.__func__.arg_checks = checks
        else:
            f.arg_checks = checks
        @wrapt.decorator
        def check_call(wrapped, instance, args, kwargs):
            try:
                config = ExecutionContext.get_context().config
                if not config.reahlsystem.runtime_checking_enabled:
                    return wrapped(*args, **kwargs)
            except (NoContextFound, AttributeError):
                pass
            return ArgumentCheckedCallable(wrapped)(*args, **kwargs)
        return check_call(f)

    return catch_wrapped


